C---------------------------------------------------------------------
C     12/09/2009  Aitbala Sargent and James Fastook
C
C      2-D full-Stokes flowline ice sheet model.  
C    
C     Parameters of the model: lambda, cx, cb, ct,
C     Parameter of the flow Lx (horizontal length scale of the domain), and 
C     Parameters of the grid: nx, nz, nt, dt 
C         - are assigned in "parameter2d.h". 
C     Bed and surface topography defined in SUBROUTINE 'testB'
C
C 
C      Exact solutions are dumped (in format: x z f) to files:
C         2du       - horizontal velocity u (in km/a, if dimensional),
C         2dw       - vertical velocity w (in km/a, if dimensional),
C         2dp       - pressure p (in kPa, if dimensional),
C         2dunorm   - sqrt(u*u+w*w) (in km/a, if dimensional),
C         2dmu      - effective viscosity (in Pa sec^2, if dimensional).
C      Compensatory stress terms are dumped to files:
C         2dsigx    - horizontal component of momentum equation
C         2dsigz    - vertical component of momentum equation
C         2dsurfx   - horizontal component of top boundary condition
C         2dsurfz   - vertical component of top boundary condition
C         2dbedx    - horizontal component of bottom boundary condition
C         2dbedz    - vertical component of bottom boundary condition
C---------------------------------------------------------------------
      IMPLICIT REAL*8(a-h,o-z)
      include "parameter2d.h"
      DIMENSION u(nx,nz),w(nx,nz),p(nx,nz)
      DIMENSION x(nx),z(nx,nz),bed(nx),surf(nx)
      DIMENSION sigx(nx,nz),sigz(nx,nz)
      DIMENSION bedx(nx),bedz(nx),surfx(nx),surfz(nx)
      DOUBLE PRECISION mu(nx,nz)
      PARAMETER(eps=1.e-20)
C ---
      CALL scales(delta,sx,sz,sp,su,sw,smu,st,ssig,ssigb)
      print *,'steady=',steady,' dimensional=',dimensional
      print *,'model parameters: cx=',cx,' cb=',cb,' ct=',ct
      print *,'domain: ',sx,' by',sz 
C ---
C --- beginning of the time loop
C ---
      DO k=1,nt
        tt=(k-1)*dt
        dx=1./(nx-1)
C ---
C --- beginning of the x- loop
C ---
        DO i=1,nx
          xx=(i-1)*dx
          CALL testB(xx,tt,b,dbdx,d2bdx2,d3bdx3,s,dsdx,d2sdx2,
     &               d3sdx3,dsdt,d2sdxdt,d3sdx2dt,dsdtint)
          x(i)=xx
          surf(i)=s
          bed(i)=b
          h=surf(i)-bed(i)
          dhdx=dsdx-dbdx
          d2hdx2=d2sdx2-d2bdx2
          d3hdx3=d3sdx3-d3bdx3
          dz=h/(nz-1)
C ---
C --- beginning of the z- loop
C ---
          DO j=1,nz
            z(i,j)=bed(i)+(j-1)*dz
            d=(surf(i)-z(i,j))/h
            dddz=-1./h
            dddx=(dsdx-dhdx*d)/h
            d2ddx2=(d2sdx2-d2hdx2*d-2.*dhdx*dddx)/h
            d3ddx3=(d3sdx3-3.*d2hdx2*dddx-3.*dhdx*d2ddx2-d*d3hdx3)/h
C ---
C --- Calculate velocities
C ---
            u(i,j)=(cx*(1-d**lambda)+cb-dsdtint)/h
            w(i,j)=u(i,j)*h*dddx+dsdt*(1-d)
C ---
C --- Calculate the first derivatives
C ---
            dudz=cx*lambda*d**(lambda-1)/h**2
            dudx=-( dhdx*u(i,j)+h**2*dudz*dddx+dsdt )/h
            dwdz=-dudx
            dwdx=dudx*h*dddx+u(i,j)*dhdx*dddx
     &      +u(i,j)*h*d2ddx2+d2sdxdt*(1.-d)-dsdt*dddx
C ---
C --- Calculate the second derivatives
C ---
            d2udz2=-cx*lambda*(lambda-1)*d**(lambda-2)/h**3
            d2udx2=-( d2hdx2*u(i,j)+2*dhdx*dudx
     &       -h**3*d2udz2*dddx**2+h**2*dudz*d2ddx2-d2sdxdt )/h
            d2udxdz=-( 2*dudx*dudz+dddx*d2udz2 )/h
            d2wdx2=d2udx2*h*dddx + 2*dudx*(dhdx*dddx+h*d2ddx2)
     &     + u(i,j)*(d2hdx2*dddx+2.*dhdx*d2ddx2+h*d3ddx3)
     &     + d3sdx2dt*(1.-d)-2.*d2sdxdt*dddx-dsdt*d2ddx2
            d2wdxdz=-d2udx2
            d2wdz2=-d2udxdz
C ---
C --- Calculate the effective viscosity and pressure:
C ---
            sxz=dudz/delta+delta*dwdx
            dsxzdx=d2udxdz/delta+delta*d2wdx2
            dsxzdz=d2udz2/delta+delta*d2wdxdz
            xnu=0.25*sxz**2-dudx*dwdz
            mu(i,j)=(xnu+eps)**(-1./3.)
            p(i,j)=2.*mu(i,j)*dudx-surf(i)+z(i,j)
C ---
C --- Calculate the compensatory stress terms:
C ---
            dmudx=(-1./3)*mu(i,j)/(xnu+eps)*
     &            (0.5*sxz*dsxzdx - d2udx2*dwdz-dudx*d2wdxdz)
            dmudz=(-1./3)*mu(i,j)/(xnu+eps)*
     &            (0.5*sxz*dsxzdz - d2udxdz*dwdz-dudx*d2wdz2)
            dpdx=2*dmudx*dudx + 2*mu(i,j)*d2udx2-dsdx
            dpdz=2*dmudz*dudx + 2*mu(i,j)*d2udxdz+1 
            sigxx=delta* (2*dmudx*dudx+2*mu(i,j)*d2udx2)
            sigxz=dmudz*sxz+mu(i,j)*dsxzdz
            sigx(i,j)=sigxx+sigxz
            sigzx=delta*(dmudx*sxz+mu(i,j)*dsxzdx)
            sigzz=2*dmudz*dwdz+2*mu(i,j)*d2wdz2
            sigz(i,j)=sigzx+sigzz
C ---
C --- Calculate the compensatory stress terms for 
C ___       the upper surface boundary conditions:
C ---
          IF(j.eq.nz) THEN
            r=delta*dsdx
            rinv=1./sqrt(1.+r*r)
            sxx=2.*mu(i,j)*dudx + p(i,j)
            szz=2.*mu(i,j)*dwdz + p(i,j)
            surfx(i)=rinv*( -r*sxx+mu(i,j)*sxz)
            surfz(i)=rinv*( -r*mu(i,j)*sxz+szz)
          ENDIF
C ---
C --- Calculate the compensatory stress terms for 
C ___       the basal boundary conditions:
C ---
          IF(j.eq.1) THEN
            r=delta*dbdx
            rinv=1./sqrt(1.+r*r)
            sxx=2.*mu(i,j)*dudx + p(i,j)
            szz=2.*mu(i,j)*dwdz + p(i,j)
            bedx(i)=rinv*( r*sxx-mu(i,j)*sxz)
            bedz(i)=rinv*( r*mu(i,j)*sxz-szz)
          ENDIF
        ENDDO     !  end of z- loop
        ENDDO     !  end of x- loop
C.................................................
        CALL dumpsurf(k,x,z,u,w,surf,bed,
     &              surfx,surfz,bedx,bedz)
C ---
C --- Dump the values at the desired time:
C ---
        IF(k.eq.1 .or. k.eq.nt) THEN
          CALL dump(k,x,z,u,w,p,mu,sigx,sigz,
     &              surfx,surfz,bedx,bedz)
        ENDIF
      ENDDO      !  end of time loop
      STOP
      END
C---------------------------------------------------------------------
      SUBROUTINE testB(x,t,b,dbdx,d2bdx2,d3bdx3,s,dsdx,d2sdx2,
     &           d3sdx3,dsdt,d2sdxdt,d3sdx2dt,dsdtint)
C---------------------------------------------------------------------
C     Return the values of the bed, surface, and their derivatives
C            for a given values of x and t.
C
C     Input:
C       x   - horizontal coordinat of the column of ice,
C       t   - time.
C---------------------------------------------------------------------
      IMPLICIT REAL*8(a-h,o-z)
      include "parameter2d.h"
        tana=tan(0.5)
        omega=8.*atan(1.)   !  2*pi
C ---
        s0x    = -tana*x
        s0xp   = -tana
        eta    = 0.5*sin(omega*x)
        etap   = 0.5*omega*cos(omega*x)
        etap2  = -0.5*omega**2*sin(omega*x)
        etap3  = -0.5*omega**3*cos(omega*x)
        etaint = -0.5/omega*cos(omega*x)
        IF(steady .or. t.eq.0.) THEN
          gamma = 0.
          gammap = 0. 
        ELSE
          gamma  = 1-exp(-ct*t)
          gammap = ct*exp(-t) 
        ENDIF
C ---
        b        = s0x-1+eta
        s        = s0x+eta*gamma
        dbdx     = s0xp+etap
        d2bdx2   = etap2
        d3bdx3   = etap3
        dsdx     = s0xp+etap*gamma
        d2sdx2   = etap2*gamma
        d3sdx3   = etap3*gamma
        dsdt     = eta*gammap
        dsdtint  = gammap*etaint
        d2sdxdt  = etap*gammap 
        d3sdx2dt = etap2*gammap 
      RETURN
      END
C---------------------------------------------------------------------
      SUBROUTINE dump(k,x,z,u,w,p,mu,sigx,sigz,
     &              surfx,surfz,bedx,bedz)
C---------------------------------------------------------------------
C     Dumps values of exact solutions and compensatory stresses 
C          at the specified time step k.
C
C     Dumps dimensional values if 'dimensional'=.true.
C           and nondimensional, otherwise.
C---------------------------------------------------------------------
      IMPLICIT REAL*8(a-h,o-z)
      INCLUDE "parameter2d.h"
      DOUBLE PRECISION mu(nx,nz)
      DIMENSION u(nx,nz),w(nx,nz),p(nx,nz)
      DIMENSION x(nx),z(nx,nz),bed(nx),surf(nx)
      DIMENSION sigx(nx,nz),sigz(nx,nz)
      DIMENSION bedx(nx),bedz(nx),surfx(nx),surfz(nx)
      CHARACTER*20 string, out(11)
      CALL scales(delta,sx,sz,sp,su,sw,smu,st,ssig,ssigb)
        IF(k.lt.10) THEN
          WRITE(unit=string,fmt='(I1)') k
        else
          WRITE(unit=string,fmt='(I2)') k
        ENDIF
        out(1)='2du'//string
        out(2)='2dw'//string
        out(3)='2dp'//string
        out(4)='2dunorm'//string
        out(5)='2dmu'//string
        out(6)='2dsigx'//string
        out(7)='2dsigz'//string
        out(8)='2dsigsurfx'//string
        out(9)='2dsigsurfz'//string
        out(10)='2dsigbedx'//string
        out(11)='2dsigbedz'//string
      DO i=21,31
        OPEN(i,file=out(i-20))
      ENDDO
        DO j=1,nz
        DO i=1,nx
          xx=x(i)
          zz=z(i,j)
          xx=x(i)*sx/1000.       ! in km
          zz=z(i,j)*sz/1000.     ! in km
          WRITE(21,*) xx,zz,u(i,j)*su
          WRITE(22,*) xx,zz,w(i,j)*sw
          WRITE(23,*) xx,zz,p(i,j)*sp
          WRITE(24,*) xx,zz,sqrt(u(i,j)*u(i,j)*su*su
     &                          +w(i,j)*w(i,j)*sw*sw)
          WRITE(25,*) xx,zz,mu(i,j)*smu
          WRITE(26,*) xx,zz,sigx(i,j)*ssig
          WRITE(27,*) xx,zz,sigz(i,j)*ssig
        ENDDO
        ENDDO
        DO i=1,nx
          xx=x(i)
          xx=x(i)*sx/1000.       ! in km
          WRITE(28,*) xx,surfx(i)*ssigb
          WRITE(29,*) xx,surfz(i)*ssigb
          WRITE(30,*) xx,bedx(i)*ssigb
          WRITE(31,*) xx,bedz(i)*ssigb
        ENDDO
        DO i=21,31
          CLOSE(i)
        ENDDO
      END
C---------------------------------------------------------------------
      SUBROUTINE dumpsurf(k,x,z,u,w,surf,bed,
     &              surfx,surfz,bedx,bedz)
C---------------------------------------------------------------------
C     Dumps values of bed and surface, as well as compensatory
C       stresses at the top and the bottom surfaces.
C       at the specified time step k.
C     Dumps dimensional values if 'dimensional'=.true.
C           and nondimensional, otherwise.
C---------------------------------------------------------------------
      IMPLICIT REAL*8(a-h,o-z)
      INCLUDE "parameter2d.h"
      DIMENSION x(nx),z(nx,nz),bed(nx),surf(nx)
      DIMENSION u(nx,nz),w(nx,nz)
      DIMENSION bedx(nx),bedz(nx),surfx(nx),surfz(nx)
      CALL scales(delta,sx,sz,sp,su,sw,smu,st,ssig,ssigb)
      IF(k.eq.1) THEN
        OPEN(41,file='2dsb.data')
        OPEN(42,file='2dsurfvel.data')
        OPEN(43,file='2dsurfx.data')
        OPEN(44,file='2dsurfz.data')
        OPEN(45,file='2dbedx.data')
        OPEN(46,file='2dbedz.data')
        DO i=1,nx
          WRITE(41,*) x(i),bed(i)*sz
        ENDDO
          WRITE(41,*) -99999,0
          WRITE(41,*) 'bed'
      ENDIF
      DO i=1,nx
        xx=x(i)
        WRITE(41,*) xx,surf(i)*sz
        WRITE(42,*) xx,sqrt(u(i,nz)*u(i,nz)*su*su
     &                          +w(i,nz)*w(i,nz)*sw*sw)
        WRITE(43,*) xx,surfx(i)*ssigb
        WRITE(44,*) xx,surfz(i)*ssigb
        WRITE(45,*) xx,bedx(i)*ssigb
        WRITE(46,*) xx,bedz(i)*ssigb
      ENDDO
      DO i=41,46
        WRITE(i,*) -99999,0
        WRITE(i,*) 't=',dt*(k-1)*st
      ENDDO
      IF(k.eq.nt) THEN
        DO i=41,46
          CLOSE(i)
        ENDDO
      ENDIF
      RETURN
      END
C---------------------------------------------------------------------
      SUBROUTINE scales(delta,sx,sz,sp,su,sw,smu,st,ssig,ssigb)
C---------------------------------------------------------------------
C     Defines scales.
C---------------------------------------------------------------------
      IMPLICIT REAL*8(a-h,o-z)
      include "parameter2d.h"
C ---
      n    = 3.                   ! Glen's parameter 
      rho  = 910.                 ! [kg/m^3]         
      g    = 9.81                 ! [m/sec^2]        
      A    = 1.e-16               ! [Pa^(-n)a^(-1)]  
      spyr = 31556926.            ! [sec/a]          
      sx    = Lx                  ! [m]              
      sz    = 1000.               ! [m]              
      delta = sz/sx               ! aspect ratio
C ---
      IF(dimensional) THEN
        sp    = rho*g*sz          ! [Pa=kg/m/sec^2]  
        su    = (2*sp)**n*sx*A    ! [m/a]           
        sw    = su*sz/sx          ! [m/a]            
        smu   = 0.5*A**(-1./n)*(su/sx)**((1.-n)/n)  ! [Pa sec^2]
        st    = sx/su             ! [a]              
        ssig  = sp/sz             ! stress scale [J=Pa/m] 
        ssigb = sp                ! bound.stress scale [Pa]  
C ---
        su    = su*0.001          !  in km/a
        sw    = sw*0.001          !  in km/a
        sp    = sp*0.000001       !  in MPa
        ssig  = ssig*0.001        !  in kJ
        ssigb = ssigb*0.001       !  in kPa
      ELSE
        sx=1.
        sz=1.
        sp=1.
        su=1.
        sw=1.
        smu=1.
        st=1.
        ssig=1.
        ssigb=1.
      ENDIF
      RETURN
      END
C---------------------------------------------------------------------

